//-*-c++-*------------------------------------------------------------
//
// File name : bioExpression.h
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Mon Apr 27 16:31:01 2009
//
//--------------------------------------------------------------------

#ifndef bioExpression_h
#define bioExpression_h

#include "patError.h"
#include <map>
#include <set>
#include <list>
#include <ostream>

#include "bioIteratorInfo.h"
#include "bioIteratorSpan.h"
#include "patVariables.h" 
#include "bioFunctionAndDerivatives.h"
class bioExpressionRepository ;
class bioArithListOfExpressions ;
class bioArithLikelihoodFctAndGrad ;
class bioArithNamedExpression ;
class bioSample ;
class trHessian ;
/*!
 This virtual class represents a generic arithmetic expression.

*/

class bioLiteral ;

class bioExpression {

  friend ostream& operator<<(ostream& str, const bioExpression& x) ;
  friend class bioExpressionRepository ;
 public:

//   typedef  enum {
//     UNARY_OP,
//     BINARY_OP,
//     LITERALS_OP,
//     RANDOM_OP,
//     CONSTANT_OP,
//     SUM_OP,
//     PROD_OP,
//     ENUM_OP,
//     MERGED_OP,
//     SEQUENCE_OP,
//     DICTIONARY_OP,
//     INTEGRAL_OP,
//     DERIVATIVE_OP,
//     UNDEFINED_OP,
//   } bioOperatorType ;


  bioExpression(bioExpressionRepository* rep, patULong par) ;

  virtual patBoolean operator==(const bioExpression& x) ;
  virtual patBoolean operator!=(const bioExpression& x) ;

  virtual ~bioExpression() ;

  /*!
    \return patTRUE if the node has no parent
  */
  virtual patBoolean isTop() const  ;
  
  /*!
    return patTRUE if the expression does not contain iterators or vectors
    bioExpression returns patTRUE. The non trivial nodes overload the function.
   */
  virtual patBoolean isSimple() const ;

  /*!  return patTRUE if the expression is a sum iterator. Typically
    used to identify a loglikelihood function.
  */
  virtual patBoolean isSumIterator() const ;

  /*!
    return patTRUE if the expression is structurally 0 so that there is no need to evaluate it. In the base class, it return patFALSE.
  */
  virtual patBoolean isStructurallyZero() const ;
  
  /*!
    return patTRUE if the expression is structurally 1. In the base class, it return patFALSE.
  */
  virtual patBoolean isStructurallyOne() const ;
  

  /*!
    \param par pointer to the parent. 
  */
  void setParent(patULong par, patError*& err) ;
  
  virtual patString check(patError*& err) const = PURE_VIRTUAL ;
  

  /**
     \return pointer to the iterator information if the top level node is an iterator, NULL otherwise. Not recursive. 
   */
  virtual patString theIterator() const ;

  /*!
    \return pointer to the parent node
  */
  virtual bioExpression* getParentPtr() const  ;
  /*!
    \return ID of the parent node
  */
  virtual patULong getParent() const  ;

  /*!
    \return name of the operator
  */
  virtual patString getOperatorName() const = PURE_VIRTUAL ;
  
  /*!
    \return value of the expression
    \param err ref. of the pointer to the error object.
  */
  virtual patReal getValue(patBoolean prepareGradient, patError*& err)  = PURE_VIRTUAL ;


  /*!
    \return value and gradient of the expression
    \param err ref. of the pointer to the error object.
  */
  virtual bioFunctionAndDerivatives* getNumericalFunctionAndGradient(vector<patULong> literalIds, patError*& err) = PURE_VIRTUAL  ;


  /*!
    \return value of the derivative w.r.t literal
    \param index of the literal involved in the derivative
    \param err ref. of the pointer to the error object.
  */
  virtual bioExpression* getDerivative(patULong aLiteralId, 
				       patError*& err) const 
    = PURE_VIRTUAL ;



  /*!
    \return A sequence of expressions containing the expression itself, followed by the derivative of it w.r.t. each literal is the list. Designed to recycle intermediary expressions.
  */
  // virtual bioArithListOfExpressions* getFunctionAndDerivatives(vector<patULong> literalIds, 
  // 							       patError*& err) const = PURE_VIRTUAL ;

  /*!
    \return In the special case of a loglikelihood function, we create an object for function and derivatives that maintain the iterator structure. 
  */
  virtual bioArithLikelihoodFctAndGrad* getLikelihoodFunctionAndDerivatives(vector<patULong> literalIds, 
									    patError*& err) const ;

  /*!
     \return printed expression
   */
  virtual patString getExpression(patError*& err) const  = PURE_VIRTUAL ;

 /*!
    Mainly for debugging
   */
  virtual patString getInfo() const ;

  /*!
     Create a deep copy of the expression and returns a pointer to it.
     It means that new instances of the children are created.
   */
  virtual bioExpression* getDeepCopy(bioExpressionRepository* rep, 
				     patError*& err) const = PURE_VIRTUAL ;
  

  /*!
     Create a shallow copy of the expression and returns a pointer to
     it.  It means that no new instance of the children are
     created. It is typically called by the repository
   */
  virtual bioExpression* getShallowCopy(bioExpressionRepository* rep, 
				     patError*& err) const = PURE_VIRTUAL ;
  
  /*!
    code of the form "prefixresult = aFormula ;"
   */
  virtual patString getCppCode(patString prefix, patError*& err) = PURE_VIRTUAL ;

  /*!
    code with only the formula
   */
  virtual patString getSimpleCppCode(patError*& err) = PURE_VIRTUAL ;

  /*!  Defines a piece of code that will be appended to the generated
    code at the end of each loop of an iterator. If the node is not an
    iterator, it does nothing. It is typically designed to compute the
    BHHH matrix.
   */
  virtual void setAdditionalCode(patString s) ;

  /*!
   */
  virtual patBoolean dependsOf(patULong aLiteralId) const  = PURE_VIRTUAL ;

  /*!
   */
  patULong getId() const;
  /*!
    The name is "bio" + Id
   */
  patString getUniqueName() const;

  /**
     A named expression is associated with a literal that can be
     reused in other expressions.  If the expression happens to be a
     constant or a literal, it does not make sense to create a named
     expresson and a NULL pointer is returned.
   */
  //virtual bioArithNamedExpression* createNamedExpression(patError*& err) const ;

  /*!
    Nameof the variable where the result of the expression in C++ is stored
   */
  patString getCppResult(patString prefix) const ;

  /*!
    Generate the code of all the GSL functions which are needed
  */
  virtual void generateGslFunctions(ostream& str_h,patString prefix,patError*& err) = PURE_VIRTUAL ;

  /*!
    Generate code for the GSL function. literalId identifies the literal corresponding to the variable of the function.
  */
  virtual patString getGslFunction(patString prefix, patULong literalId, patULong parentId, patError*& err)  ;

  /*!
    Declaration of the variable where the result of the expression in C++ is stored
   */
  patString getCppResultDeclaration(patString prefix) const ;

  /*!
    Force the generation of the cpp code for iterators
   */
  virtual void forceCodeGeneration()  = PURE_VIRTUAL ;

  /*!
   */
  virtual patBoolean containsAnIterator() const  = PURE_VIRTUAL ;

  /*!
   */
  virtual patBoolean containsAnIntegral() const  = PURE_VIRTUAL ;

  /*!
   */
  virtual patBoolean containsASequence() const  = PURE_VIRTUAL ;


  /*
    Count the number of operations
   */
  virtual patULong getNumberOfOperations() const  = PURE_VIRTUAL ;

  /*
    If an expression generates code requiring some include file, this function generates the include statements.
   */
  virtual patString getIncludeStatements(patString prefix) const = PURE_VIRTUAL ;


  /*!
   */
  virtual void simplifyZeros(patError*& err) = PURE_VIRTUAL ;

  /**
   */
  virtual patBoolean isConstant() const ;

  /**
   */
  virtual patBoolean isSimulator() const ;

  /**
   */
  virtual patBoolean isSequence() const ;
  /**
   */
  virtual patBoolean isLiteral() const ;

  virtual void setVariables(const patVariables* x) = PURE_VIRTUAL ;

  virtual void setDraws(const vector< vector<patReal> >* d) = PURE_VIRTUAL ;

  virtual void setSample(const bioSample* s) = PURE_VIRTUAL;

  virtual void setCurrentSpan(bioIteratorSpan aSpan) = PURE_VIRTUAL ;

  virtual void setThreadSpan(bioIteratorSpan aSpan) = PURE_VIRTUAL ;

  // This function returns a non null pointer only for the top sum iterator.
  virtual trHessian* getBhhh() ;

  patULong getThreadId() const ;
  /**
     Compute a string that represents the expression. It is designed to replace the expression itself when used only for comparison purposes. 
     Code:
     +{expr1}{expr2}: binary plus
     -{expr1}{expr2}: binary minus
     *{expr1}{expr2}: multiplication
     /{expr1}{expr2}: division
     ^{expr1}{expr2}: power
     &{expr1}{expr2}: and
     |{expr1}{expr2}: or
     ={expr1}{expr2}: equal
     !={expr1}{expr2}: not equal
     <{expr1}{expr2}: lesser than
     <={expr1}{expr2}: lesser or equal to
     >{expr1}{expr2}: greater than
     >={expr1}{expr2}: greater or equal to 
     $A{expr}: abs
     $D[expr][{expr1}...{exprN}]: dictionary (bioArithElem)
     $E{expr}: exp
     $L{expr}: log
     $M{expr}: Unary minus
     $Piterator_name{expr}: prod
     $Q{string1}{string2}: sequence
     $Siterator_name{expr}: sum
     $Ziterator_name[{expr1}...{exprN}]: merged sum
     //{expr1}{expr2}...{exprN}//: list of expressions
     number: constant
     #id: literal
     &id: random
   */
  virtual patString getExpressionString() const = PURE_VIRTUAL ;

  virtual void collectExpressionIds(set<patULong>* s) const = PURE_VIRTUAL ;

  virtual patBoolean isNamedExpression() const ;


protected:

  patULong parent ;

  patULong theId ;
  patBoolean codeMustBeGenerated ;
  patString additionalCode ;

  const patVariables* __x ;
  const vector< vector<patReal> >* __draws ;
  const bioSample* theSample ;

  bioIteratorSpan theCurrentSpan ;
  bioIteratorSpan theThreadSpan ;

  //  vector<patULong> theGradientLiteralIds ;

  // These are the expressions such that we need to propagate
  // information such as threadId or the data sample
  vector<bioExpression*> relatedExpressions ;

  bioExpressionRepository* theRepository ;

  bioFunctionAndDerivatives result ;
};


#endif
